program WaveDMA;
{****************************************************************************
** Demonstration of playing a single wave file using DMA                   **
**  by Steven H Don                                                        **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************}

uses Dos, Crt;

type WaveData = record
  SoundLength,
  Frequency   : Word;
  Sample      : Pointer;
end;

var
  Voice     : WaveData; {Pointer to wave file}

  Base      : Word;     {Sound Blaster base address}

  WaveFile  : String;   {File name for the wave file to be played}

{****************************************************************************
** Checks to see if a Sound Blaster exists at a given address, returns     **
** true if Sound Blaster found, false if not.                              **
****************************************************************************}
function ResetDSP(Test : Word) : Boolean;
begin
  {Reset the DSP}
  Port[Test + $6] := 1;
  Delay(10);
  Port[Test + $6] := 0;
  Delay(10);
  {Check if reset was succesfull}
  if (Port[Test + $E] and $80 = $80) and (Port[Test + $A] = $AA) then begin
    {DSP was found}
    ResetDSP := true;
    Base := Test;
  end else
    {No DSP was found}
    ResetDSP := false;
end;

{****************************************************************************
** Send a byte to the DSP (Digital Signal Processor) on the Sound Blaster  **
****************************************************************************}
procedure WriteDSP(Value : byte);
begin
  {Wait for the DSP to be ready to accept data}
  while Port[Base + $C] and $80 <> 0 do;
  {Send byte}
  Port[Base + $C] := Value;
end;

{****************************************************************************
** Plays a part of the memory                                              **
****************************************************************************}
procedure PlayBack (Wave : WaveData);
var
  LinearAddress : LongInt;
  Page, OffSet  : Word;
  TimeConstant  : Byte;

begin
  with Wave do begin
    TimeConstant := (65536 - (256000000 div Frequency)) shr 8;
    WriteDSP($40);                {DSP-command 40h - Set sample frequency}
    WriteDSP(TimeConstant);       {Write time constant for 11025Hz}

    {Convert pointer to linear address}
    LinearAddress := Seg (Wave.Sample^);
    LinearAddress := LinearAddress shl 4 + Ofs (Wave.Sample^);
    Page := LinearAddress shr 16;  {Calculate page}
    OffSet := LinearAddress and $FFFF; {Calculate offset in the page}
    {
      Note - this procedure only works with DMA channel 1
    }
    Port[$0A] := 5;               {Mask DMA channel 1}
    Port[$0C] := 0;               {Clear byte pointer}
    Port[$0B] := $49;             {Set mode}
    {
      The mode consists of the following:
      $49 = binary 01 00 10 01
                   |  |  |  |
                   |  |  |  +- DMA channel 01
                   |  |  +---- Read operation (the DSP reads from memory)
                   |  +------- Single cycle mode
                   +---------- Block mode
    }
    Port[$02] := Lo(OffSet);      {Write the offset to the DMA controller}
    Port[$02] := Hi(OffSet);

    Port[$83] := Page;            {Write the page to the DMA controller}

    Port[$03] := lo(SoundLength); {Set the block length to length of the wave}
    Port[$03] := hi(SoundLength);

    Port[$0A] := 1;               {Unmask DMA channel}

    WriteDSP($14);                {DSP-command 14h - Single cycle playback}
    WriteDSP(lo(SoundLength));
    WriteDSP(hi(SoundLength));
  end;
end;

{****************************************************************************
** Loads a wave file into memory.                                          **
** This procedure expects a _very_ standard wave header.                   **
** It doesn't perform much error checking.                                 **
****************************************************************************}
function LoadVoice (var Voice : WaveData; FileName : String) : Boolean;
var
  WAVFile : File;

  Header : record
    RIFF      : LongInt;                  {RIFF header}
    NI1       : Array [$04..$07] of Byte; {not important}
    WAVE      : LongInt;                  {WAVE identifier}
    fmt       : LongInt;                  {fmt  identifier}
    NI2       : Array [$10..$15] of Byte; {not important}
    Channels  : Word;                     {channels 1 = mono; 2 = stereo}
    Frequency : LongInt;                  {sample frequency}
    NI3       : Array [$1C..$21] of Byte; {not important}
    BitRes    : Word;                     {bit resolution 8/16 bit}
    data      : LongInt;                  {data identifier}
    datasize  : LongInt;                  {size of data chunk}
  end;

begin
  LoadVoice := true;
  {$I-}
  FileMode := 0; {Read only}
  Assign (WAVFile, FileName); {Open the file}
  Reset (WAVFile, 1);
  {If it can't be opened...}
  if IOResult <> 0 then begin
    {..display error message}
    Writeln ('Unable to open wave file');
    LoadVoice := false;
    Exit;
  end;
  {$I+}

  {Load the sample data}
  BlockRead (WAVFile, Header, sizeof (Header));

  {Check RIFF, WAVE, fmt and data headers}
  if (Header.RIFF <> $46464952) or
     (Header.WAVE <> $45564157) or
     (Header.fmt  <> $20746D66) or
     (Header.data <> $61746164) then begin
    Writeln ('Not a wave file');
    LoadVoice := false;
    Exit;
  end;

  {Check channels}
  if Header.Channels <> 1 then begin
    Writeln ('Not a mono wave file');
    LoadVoice := false;
    Exit;
  end;

  {Check bit resolution}
  if Header.BitRes <> 8 then begin
    Writeln ('Not an 8-bit wave file');
    LoadVoice := false;
    Exit;
  end;

  {Return length of file}
  Voice.SoundLength := Header.datasize;

  {Check size}
  if Voice.SoundLength > 32000 then begin
    Writeln ('Warning: sample size is more than 32K, this may cause problems with EMM386');
    Writeln ('Press ESCAPE to quit, or any other key to ignore.');
    if ReadKey = Chr (27) then begin
      LoadVoice := false;
      Exit;
    end;
    if Voice.SoundLength > 64000 then begin
      Writeln ('File to large, only first part will load and play.');
      Voice.SoundLength := 64000;
    end;
  end;

  Voice.Frequency := Header.Frequency;

  {Load the sample data}
  GetMem (Voice.Sample, Voice.SoundLength); {Assign memory}
  BlockRead (WAVFile, Voice.Sample^, Voice.SoundLength);

  Close (WAVFile); {Close the file}
end;

begin
  {Clear screen}
  ClrScr;
  Writeln ('Playing a wave file using DMA - by Steven Don');
  Writeln ('---------------------------------------------');

  {Check for Sound Blaster}
  if ResetDSP ($220) then begin
    {at 220h}
    Writeln ('Sound Blaster found at 220h');
  end else begin
    if ResetDSP ($240) then begin
      {at 240h}
      Writeln ('Sound Blaster found at 240h');
    end else begin
      {or none at all}
      Writeln ('No Sound Blaster found');
      Halt;
    end;
  end;

  Write ('Please input file name + extension: ');
  Readln (WaveFile);

  {Load wave file}
  if LoadVoice (Voice, WaveFile) then begin
    {Start playback}
    PlayBack (Voice);

    Writeln ('Press any key to exit...');
    {Wait for a keypress}
    ReadKey;

    {Stops DMA-transfer}
    WriteDSP ($D0);
  end;
end.